package com.me.common.ecs;

import com.badlogic.gdx.utils.Array;

import java.util.HashMap;
import java.util.Map;

public abstract class EntitySystem extends BaseSystem {

    private static final Map<Class<? extends EntitySystem>, Long> systemBits = new HashMap<>();
    private static long nextBit = 1l;

    private long systemBit;
    private long typeBits;

    protected Array<Entity> entities;

    public EntitySystem(Engine engine, Class<? extends Component>... components) {
        super(engine);
        this.systemBit = getBitFor(getClass());
        this.typeBits = ComponentType.getMaskBits(components);
        this.entities = new Array<>(true, 16, Entity.class);
    }

    public Array<Entity> getEntities() {
        return entities;
    }

    public void process(float dt) {
        for (int i = 0, n = entities.size; i < n; i++) {
            Entity entity = entities.items[i];
            if (!entity.removed && entity.active) {
                processEntity(entity, dt);
            }
        }
    }

    public abstract void processEntity(Entity entity, float dt);

    protected void refresh(Entity entity) {
        boolean enabled = (entity.systemEnabledBits & systemBit) == systemBit;
        boolean interested = (entity.componentBits & typeBits) == typeBits;

        if (interested && !enabled && entity.active && !entity.removed) {
            add(entity);
        } else if (enabled && (!interested || !entity.active || entity.removed)) {
            remove(entity);
        }
    }

    private void add(Entity entity) {
        entity.addSystemEnabledBit(systemBit);
        entities.add(entity);
    }

    private void remove(Entity entity) {
        entity.removeSystemEnabledBit(systemBit);
        entities.removeValue(entity, true);
    }


    static long getBitFor(Class<? extends EntitySystem> es) {
        Long bits = systemBits.get(es);
        if (bits == null) {
            bits = nextBit;
            nextBit <<= 1;
            systemBits.put(es, bits);
        }
        return bits;
    }

}
